from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Sort(_message.Message):
    __slots__ = ["key", "direction"]
    class Direction(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        DESCENDING: _ClassVar[Sort.Direction]
        ASCENDING: _ClassVar[Sort.Direction]
    DESCENDING: Sort.Direction
    ASCENDING: Sort.Direction
    KEY_FIELD_NUMBER: _ClassVar[int]
    DIRECTION_FIELD_NUMBER: _ClassVar[int]
    key: str
    direction: Sort.Direction
    def __init__(self, key: _Optional[str] = ..., direction: _Optional[_Union[Sort.Direction, str]] = ...) -> None: ...

class ListRequest(_message.Message):
    __slots__ = ["limit", "token", "sort_by", "filters", "raw_filters", "sort_by_fields"]
    LIMIT_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    SORT_BY_FIELD_NUMBER: _ClassVar[int]
    FILTERS_FIELD_NUMBER: _ClassVar[int]
    RAW_FILTERS_FIELD_NUMBER: _ClassVar[int]
    SORT_BY_FIELDS_FIELD_NUMBER: _ClassVar[int]
    limit: int
    token: str
    sort_by: Sort
    filters: _containers.RepeatedCompositeFieldContainer[Filter]
    raw_filters: _containers.RepeatedScalarFieldContainer[str]
    sort_by_fields: _containers.RepeatedCompositeFieldContainer[Sort]
    def __init__(self, limit: _Optional[int] = ..., token: _Optional[str] = ..., sort_by: _Optional[_Union[Sort, _Mapping]] = ..., filters: _Optional[_Iterable[_Union[Filter, _Mapping]]] = ..., raw_filters: _Optional[_Iterable[str]] = ..., sort_by_fields: _Optional[_Iterable[_Union[Sort, _Mapping]]] = ...) -> None: ...

class Filter(_message.Message):
    __slots__ = ["function", "field", "values"]
    class Function(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        EQUAL: _ClassVar[Filter.Function]
        NOT_EQUAL: _ClassVar[Filter.Function]
        GREATER_THAN: _ClassVar[Filter.Function]
        GREATER_THAN_OR_EQUAL: _ClassVar[Filter.Function]
        LESS_THAN: _ClassVar[Filter.Function]
        LESS_THAN_OR_EQUAL: _ClassVar[Filter.Function]
        CONTAINS: _ClassVar[Filter.Function]
        VALUE_IN: _ClassVar[Filter.Function]
        ENDS_WITH: _ClassVar[Filter.Function]
        NOT_ENDS_WITH: _ClassVar[Filter.Function]
    EQUAL: Filter.Function
    NOT_EQUAL: Filter.Function
    GREATER_THAN: Filter.Function
    GREATER_THAN_OR_EQUAL: Filter.Function
    LESS_THAN: Filter.Function
    LESS_THAN_OR_EQUAL: Filter.Function
    CONTAINS: Filter.Function
    VALUE_IN: Filter.Function
    ENDS_WITH: Filter.Function
    NOT_ENDS_WITH: Filter.Function
    FUNCTION_FIELD_NUMBER: _ClassVar[int]
    FIELD_FIELD_NUMBER: _ClassVar[int]
    VALUES_FIELD_NUMBER: _ClassVar[int]
    function: Filter.Function
    field: str
    values: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, function: _Optional[_Union[Filter.Function, str]] = ..., field: _Optional[str] = ..., values: _Optional[_Iterable[str]] = ...) -> None: ...



from __future__ import annotations

import datetime as _datetime
import hashlib
import logging
import os
import re
from copy import deepcopy
from dataclasses import asdict, dataclass
from functools import partial, wraps
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from flytekit import PodTemplate
from flytekit.configuration import SerializationSettings
from flytekit.core import launch_plan as _annotated_launchplan
from flytekit.core import workflow as _annotated_workflow
from flytekit.core.base_task import TaskResolverMixin
from flytekit.core.python_auto_container import get_registerable_container_image
from flytekit.core.python_function_task import PythonFunctionTask
from flytekit.core.resources import Resources
from flytekit.core.task import TaskPlugins
from flytekit.core.task import task as flytekit_task_decorator
from flytekit.extras.accelerators import BaseAccelerator
from flytekit.image_spec import ImageSpec
from flytekit.models.documentation import Documentation
from flytekit.models.security import Secret

from union.actor._hash import _make_key

logger = logging.getLogger(__name__)


_ACTOR_NAME_RE = re.compile(r"[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*")
_ACTOR_STATE_CACHE = {}


def actor_cache(f):
    """Cache function between actor executions."""

    @wraps(f)
    def inner_func(*args, **kwargs):
        args_with_extra = (os.getcwd(), f.__name__, f.__module__, *args)
        key = _make_key(args_with_extra, kwargs)
        try:
            return _ACTOR_STATE_CACHE[key]
        except KeyError:
            _ACTOR_STATE_CACHE[key] = f(*args, **kwargs)
            return _ACTOR_STATE_CACHE[key]

    return inner_func


@dataclass
class ActorEnvironment:
    """
    ActorEnvironment class.

    :param name: The name of the actor. This is used in conjunction with the
        project, domain, and version to uniquely identify the actor.
    :param container_image: The container image to use for the task. Set to
        default image if none provided.
    :param replica_count: The number of workers to provision that are able to
        accept tasks.
    :param ttl_seconds: How long to keep the Actor alive while no tasks are
        being run. If not provided the default configuration value of 90s
        will be used.
    :param environment: Environment variables as key, value pairs in a Python
        dictionary.
    :param requests: Compute resource requests per task.
    :param limits: Compute resource limits.
    :param accelerator: The accelerator device to use for the task.
    :param secret_requests: Keys (ideally descriptive) that can identify the
        secrets supplied at runtime.
    :param pod_template: The pod template to use as the base configuration
        for actor replica pods.
    :param interruptible: Whether the actor replica pods are labelled as
        interruptible.
    """

    # These fields here are part of the actor config.
    name: str
    container_image: Optional[Union[str, ImageSpec]] = None
    replica_count: int = 1
    ttl_seconds: Optional[int] = None
    environment: Optional[Dict[str, str]] = None
    requests: Optional[Resources] = None
    limits: Optional[Resources] = None
    accelerator: Optional[BaseAccelerator] = None
    secret_requests: Optional[List[Secret]] = None
    pod_template: Optional[PodTemplate] = None
    interruptible: bool = None

    def __post_init__(self):
        match_re = _ACTOR_NAME_RE.fullmatch(self.name)
        if not match_re:
            raise ValueError("Actor name must consist of lower case alphanumeric characters, '-' or '.'")

        if self.ttl_seconds is not None and (self.ttl_seconds < 30 or self.ttl_seconds > 900):
            raise ValueError("Actor TTL must be between 30s and 15m (900s)")

    def _task(
        self,
        _task_function=None,
        cache: bool = False,
        cache_serialize: bool = False,
        cache_version: str = "",
        retries: int = 0,
        timeout: Union[_datetime.timedelta, int] = 0,
        node_dependency_hints: Optional[
            Iterable[
                Union[
                    PythonFunctionTask,
                    _annotated_launchplan.LaunchPlan,
                    _annotated_workflow.WorkflowBase,
                ]
            ]
        ] = None,
        task_resolver: Optional[TaskResolverMixin] = None,
        docs: Optional[Documentation] = None,
        execution_mode: PythonFunctionTask.ExecutionBehavior = PythonFunctionTask.ExecutionBehavior.DEFAULT,
        **kwargs,
    ):
        wrapper = partial(
            flytekit_task_decorator,
            task_config=self,
            cache=cache,
            cache_serialize=cache_serialize,
            cache_version=cache_version,
            retries=retries,
            timeout=timeout,
            node_dependency_hints=node_dependency_hints,
            task_resolver=task_resolver,
            docs=docs,
            container_image=self.container_image,
            environment=self.environment,
            requests=self.requests,
            limits=self.limits,
            accelerator=self.accelerator,
            secret_requests=self.secret_requests,
            execution_mode=execution_mode,
            pod_template=self.pod_template,
            interruptible=self.interruptible,
            **kwargs,
        )

        if _task_function:
            return wrapper(_task_function=_task_function)
        return wrapper

    @property
    def task(self):
        return self._task

    # Keeping this private for now to test
    @property
    def _dynamic(self):
        return partial(self._task, execution_mode=PythonFunctionTask.ExecutionBehavior.DYNAMIC)

    @property
    def version(self) -> str:
        fields = []
        for k, v in asdict(self).items():
            if isinstance(v, BaseAccelerator):
                fields.append(str(v.to_flyte_idl()))
            elif k == "pod_template":
                # Remove args from the version specification
                try:
                    for container in v["pod_spec"].containers:
                        container.args = []
                except Exception:
                    pass
                fields.append(str(v))
            else:
                fields.append(str(v))
        all_fields = "".join(fields)
        hash_object = hashlib.md5(all_fields.encode("utf-8"))
        hex_digest = hash_object.hexdigest()
        logger.debug(f"Computing actor hash with string {all_fields} to {hex_digest[:15]}")
        return hex_digest[:15]


class ActorTask(PythonFunctionTask[ActorEnvironment]):
    _ACTOR_TASK_TYPE = "actor"

    def __init__(self, task_config: ActorEnvironment, task_function: Callable, **kwargs):
        super(ActorTask, self).__init__(
            task_config=task_config,
            task_type=self._ACTOR_TASK_TYPE,
            task_function=task_function,
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Optional[Dict[str, Any]]:
        """
        Serialize the `ActorTask` config into a dict.

        :param settings: Current serialization settings
        :return: Dictionary representation of the dask task config.
        """
        # Get version with the actual container_image value
        container_image = get_registerable_container_image(self.container_image, settings.image_config)
        task_config = deepcopy(self.task_config)
        task_config.container_image = container_image

        return {
            "name": self.task_config.name,
            "version": task_config.version,
            "type": self._ACTOR_TASK_TYPE,
            "spec": {
                "container_image": container_image,
                "backlog_length": None,
                "parallelism": 1,
                "replica_count": self.task_config.replica_count,
                "ttl_seconds": self.task_config.ttl_seconds,
            },
        }

    @classmethod
    def from_task(
        cls,
        task: PythonFunctionTask,
        replica_count: int,
        ttl_seconds: int = 60,
        name: Optional[str] = None,
    ) -> "ActorTask":
        """
        Create an ActorTask from a PythonFunctionTask.
        """

        if not isinstance(task, PythonFunctionTask):
            raise ValueError("Only PythonFunctionTask can be converted to ActorTask")

        if name is None:
            name = task.task_function.__name__.replace("_", "-")

        actor = ActorEnvironment(
            container_image=task.container_image,
            name=name,
            replica_count=replica_count,
            ttl_seconds=ttl_seconds,
            environment=task.environment,
            requests=task.resources.requests,
            limits=task.resources.limits,
            secret_requests=task._security_ctx.secrets if task._security_ctx else None,
            accelerator=task.accelerator,
            pod_template=task.pod_template,
        )

        return cls(
            task_config=actor,
            task_function=task.task_function,
            retries=task._metadata.retries,
            timeout=task._metadata.timeout,
            cache=task._metadata.cache,
            cache_serialize=task._metadata.cache_serialize,
            cache_version=task._metadata.cache_version,
            docs=task.docs,
            execution_mode=task.execution_mode,
        )


TaskPlugins.register_pythontask_plugin(ActorEnvironment, ActorTask)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common/identity.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.common import identifier_pb2 as common_dot_identifier__pb2
from union.internal.common import policy_pb2 as common_dot_policy__pb2
from union.internal.common import role_pb2 as common_dot_role__pb2
from union.internal.validate.validate import validate_pb2 as validate_dot_validate__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x63ommon/identity.proto\x12\x0f\x63loudidl.common\x1a\x17\x63ommon/identifier.proto\x1a\x13\x63ommon/policy.proto\x1a\x11\x63ommon/role.proto\x1a\x17validate/validate.proto\"\xcc\x01\n\x04User\x12/\n\x02id\x18\x01 \x01(\x0b\x32\x1f.cloudidl.common.UserIdentifierR\x02id\x12-\n\x04spec\x18\x02 \x01(\x0b\x32\x19.cloudidl.common.UserSpecR\x04spec\x12/\n\x05roles\x18\x03 \x03(\x0b\x32\x15.cloudidl.common.RoleB\x02\x18\x01R\x05roles\x12\x33\n\x08policies\x18\x04 \x03(\x0b\x32\x17.cloudidl.common.PolicyR\x08policies\"\xd6\x01\n\x08UserSpec\x12\x1d\n\nfirst_name\x18\x01 \x01(\tR\tfirstName\x12\x1b\n\tlast_name\x18\x02 \x01(\tR\x08lastName\x12\x14\n\x05\x65mail\x18\x03 \x01(\tR\x05\x65mail\x12\"\n\x0corganization\x18\x04 \x01(\tR\x0corganization\x12\x1f\n\x0buser_handle\x18\x05 \x01(\tR\nuserHandle\x12\x16\n\x06groups\x18\x06 \x03(\tR\x06groups\x12\x1b\n\tphoto_url\x18\x07 \x01(\tR\x08photoUrl\"s\n\x0b\x41pplication\x12\x36\n\x02id\x18\x01 \x01(\x0b\x32&.cloudidl.common.ApplicationIdentifierR\x02id\x12,\n\x04spec\x18\x02 \x01(\x0b\x32\x18.cloudidl.common.AppSpecR\x04spec\"A\n\x07\x41ppSpec\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\"\n\x0corganization\x18\x02 \x01(\tR\x0corganization\"\xa7\x01\n\x10\x45nrichedIdentity\x12\x35\n\x04user\x18\x01 \x01(\x0b\x32\x15.cloudidl.common.UserB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\x04user\x12J\n\x0b\x61pplication\x18\x02 \x01(\x0b\x32\x1c.cloudidl.common.ApplicationB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01H\x00R\x0b\x61pplicationB\x10\n\tprincipal\x12\x03\xf8\x42\x01\"\xa4\x01\n\x08Identity\x12:\n\x07user_id\x18\x01 \x01(\x0b\x32\x1f.cloudidl.common.UserIdentifierH\x00R\x06userId\x12O\n\x0e\x61pplication_id\x18\x02 \x01(\x0b\x32&.cloudidl.common.ApplicationIdentifierH\x00R\rapplicationIdB\x0b\n\tprincipalB\xae\x01\n\x13\x63om.cloudidl.commonB\rIdentityProtoH\x02P\x01Z)github.com/unionai/cloud/gen/pb-go/common\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43loudidl.Common\xca\x02\x0f\x43loudidl\\Common\xe2\x02\x1b\x43loudidl\\Common\\GPBMetadata\xea\x02\x10\x43loudidl::Commonb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common.identity_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\023com.cloudidl.commonB\rIdentityProtoH\002P\001Z)github.com/unionai/cloud/gen/pb-go/common\242\002\003CCX\252\002\017Cloudidl.Common\312\002\017Cloudidl\\Common\342\002\033Cloudidl\\Common\\GPBMetadata\352\002\020Cloudidl::Common'
  _USER.fields_by_name['roles']._options = None
  _USER.fields_by_name['roles']._serialized_options = b'\030\001'
  _ENRICHEDIDENTITY.oneofs_by_name['principal']._options = None
  _ENRICHEDIDENTITY.oneofs_by_name['principal']._serialized_options = b'\370B\001'
  _ENRICHEDIDENTITY.fields_by_name['user']._options = None
  _ENRICHEDIDENTITY.fields_by_name['user']._serialized_options = b'\372B\005\212\001\002\020\001'
  _ENRICHEDIDENTITY.fields_by_name['application']._options = None
  _ENRICHEDIDENTITY.fields_by_name['application']._serialized_options = b'\372B\005\212\001\002\020\001'
  _globals['_USER']._serialized_start=133
  _globals['_USER']._serialized_end=337
  _globals['_USERSPEC']._serialized_start=340
  _globals['_USERSPEC']._serialized_end=554
  _globals['_APPLICATION']._serialized_start=556
  _globals['_APPLICATION']._serialized_end=671
  _globals['_APPSPEC']._serialized_start=673
  _globals['_APPSPEC']._serialized_end=738
  _globals['_ENRICHEDIDENTITY']._serialized_start=741
  _globals['_ENRICHEDIDENTITY']._serialized_end=908
  _globals['_IDENTITY']._serialized_start=911
  _globals['_IDENTITY']._serialized_end=1075
# @@protoc_insertion_point(module_scope)



import typing
from logging import getLogger
from typing import AsyncIterable, Callable

import grpc
from flytekit.clients.auth.authenticator import Authenticator

from union._interceptor import generate_random_str

DEFAULT_KUBERNETES_TOKEN_PATH = "/var/run/secrets/kubernetes.io/serviceaccount/token"

logger = getLogger(__name__)


# See https://github.com/grpc/grpc/issues/34537 and proposed fix https://github.com/grpc/grpc/pull/36487
# Unfortunately we can't use one interceptor for unary and streaming endpoints
class ServiceAccountMetadataInterceptor:
    """
    Defines middleware to attach the service account token as authorization metadata for clients calling object store
    in an execution context
    """

    def __init__(self):
        self._service_account_token = self._read_service_account_token()

    def _read_service_account_token(self) -> typing.Optional[None]:
        try:
            with open(DEFAULT_KUBERNETES_TOKEN_PATH, "r") as token_file:
                token = token_file.read()
            return token
        except Exception as e:
            logger.debug(f"Error reading service account token: {e}, not adding to request")
            return None

    def _inject_default_metadata(self, client_call_details: grpc.aio.ClientCallDetails) -> grpc.aio.ClientCallDetails:
        if self._service_account_token is None:
            return client_call_details

        if client_call_details.metadata is not None:
            old_metadata = list(client_call_details.metadata)
        else:
            old_metadata = []

        new_metadata = [*old_metadata, ("authorization", f"Bearer {self._service_account_token}")]

        new_details = grpc.aio.ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=new_metadata,
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
        )
        return new_details


class UnaryUnaryClientServiceAccountMetadataInterceptor(
    grpc.aio.UnaryUnaryClientInterceptor, ServiceAccountMetadataInterceptor
):
    async def intercept_unary_unary(
        self, continuation: Callable, client_call_details: grpc.aio.ClientCallDetails, request: typing.Any
    ) -> typing.Any:
        new_client_call_details = super()._inject_default_metadata(client_call_details)
        return await continuation(new_client_call_details, request)


class UnaryStreamClientServiceAccountMetadataInterceptor(
    grpc.aio.UnaryStreamClientInterceptor, ServiceAccountMetadataInterceptor
):
    async def intercept_unary_stream(
        self, continuation: Callable, client_call_details: grpc.aio.ClientCallDetails, request: any
    ) -> AsyncIterable:
        new_client_call_details = super()._inject_default_metadata(client_call_details)
        return await continuation(new_client_call_details, request)


class StreamUnaryClientServiceAccountMetadataInterceptor(
    grpc.aio.StreamUnaryClientInterceptor, ServiceAccountMetadataInterceptor
):
    async def intercept_stream_unary(
        self,
        continuation: Callable,
        client_call_details: grpc.aio.ClientCallDetails,
        request_iterator: any,
    ) -> grpc.aio.StreamUnaryCall:
        new_client_call_details = super()._inject_default_metadata(client_call_details)
        return await continuation(new_client_call_details, request_iterator)


class StreamStreamClientServiceAccountMetadataInterceptor(
    grpc.aio.StreamStreamClientInterceptor, ServiceAccountMetadataInterceptor
):
    async def intercept_stream_stream(
        self,
        continuation: Callable,
        client_call_details: grpc.aio.ClientCallDetails,
        request_iterator: typing.AsyncIterator[any],
    ) -> grpc.aio.Call:
        new_client_call_details = super()._inject_default_metadata(client_call_details)
        return await continuation(new_client_call_details, request_iterator)


class AuthInterceptor:
    def __init__(self, authenticator: Authenticator):
        self._authenticator = authenticator

    async def _call_details_with_auth_metadata(
        self, client_call_details: grpc.aio.ClientCallDetails
    ) -> grpc.aio.ClientCallDetails:
        """
        Returns new ClientCallDetails with metadata added.
        """
        request_id = f"u-{generate_random_str(20)}"
        metadata = []
        if client_call_details.metadata:
            metadata.extend(list(client_call_details.metadata))
        metadata.append(("x-request-id", request_id))
        auth_metadata = self._authenticator.fetch_grpc_call_auth_metadata()
        if auth_metadata:
            metadata.append(auth_metadata)

        return grpc.aio.ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=metadata,
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
        )


class UnaryUnaryClientAuthInterceptor(grpc.aio.UnaryUnaryClientInterceptor, AuthInterceptor):
    async def intercept_unary_unary(
        self, continuation: Callable, client_call_details: grpc.aio.ClientCallDetails, request: typing.Any
    ) -> typing.Any:
        new_client_call_details = await super()._call_details_with_auth_metadata(client_call_details)
        try:
            continuation = await continuation(new_client_call_details, request)
            return continuation
        except Exception as e:
            if not hasattr(e, "code"):
                raise e
            if e.code() == grpc.StatusCode.UNAUTHENTICATED or e.code() == grpc.StatusCode.UNKNOWN:
                self._authenticator.refresh_credentials()
                updated_call_details = await super()._call_details_with_auth_metadata(client_call_details)
                return continuation(updated_call_details, request)


class UnaryStreamClientAuthInterceptor(grpc.aio.UnaryStreamClientInterceptor, AuthInterceptor):
    async def intercept_unary_stream(
        self, continuation: Callable, client_call_details: grpc.aio.ClientCallDetails, request: any
    ) -> AsyncIterable:
        new_client_call_details = await super()._call_details_with_auth_metadata(client_call_details)
        return await continuation(new_client_call_details, request)


class StreamUnaryClientAuthInterceptor(grpc.aio.StreamUnaryClientInterceptor, AuthInterceptor):
    async def intercept_stream_unary(
        self,
        continuation: Callable,
        client_call_details: grpc.aio.ClientCallDetails,
        request_iterator: any,
    ) -> grpc.aio.StreamUnaryCall:
        new_client_call_details = await super()._call_details_with_auth_metadata(client_call_details)
        return await continuation(new_client_call_details, request_iterator)


class StreamStreamClientAuthInterceptor(grpc.aio.StreamStreamClientInterceptor, AuthInterceptor):
    async def intercept_stream_stream(
        self,
        continuation: Callable,
        client_call_details: grpc.aio.ClientCallDetails,
        request_iterator: typing.AsyncIterator[any],
    ) -> grpc.aio.Call:
        new_client_call_details = await super()._call_details_with_auth_metadata(client_call_details)
        return await continuation(new_client_call_details, request_iterator)



from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ImageIdentifier(_message.Message):
    __slots__ = ["name"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class Image(_message.Message):
    __slots__ = ["id", "fqin"]
    ID_FIELD_NUMBER: _ClassVar[int]
    FQIN_FIELD_NUMBER: _ClassVar[int]
    id: ImageIdentifier
    fqin: str
    def __init__(self, id: _Optional[_Union[ImageIdentifier, _Mapping]] = ..., fqin: _Optional[str] = ...) -> None: ...



import logging
import os
import sys
from typing import Generator

import sglang.srt.model_loader.loader
import torch
from sglang.srt.configs.device_config import DeviceConfig
from sglang.srt.configs.model_config import ModelConfig
from sglang.srt.server_args import prepare_server_args
from sglang.srt.utils import kill_process_tree

try:
    from sglang.srt.server import launch_server
except ImportError:
    from sglang import launch_server


from ..config import (
    LOCAL_MODEL_PATH,
    REMOTE_MODEL_PATH,
    STREAM_SAFETENSORS,
)
from ..loader import SafeTensorsStreamer, prefetch, prefix_exists

logger = logging.getLogger(__name__)

_OrigDefaultModelLoader = sglang.srt.model_loader.loader.DefaultModelLoader


class UnionModelLoader(_OrigDefaultModelLoader):
    def _get_weights_iterator(self, source) -> Generator[tuple[str, torch.Tensor], None, None]:
        # Try to load weights using the Union SafeTensorsStreamer. Fallback to the default loader otherwise.
        try:
            streamer = SafeTensorsStreamer(REMOTE_MODEL_PATH, LOCAL_MODEL_PATH)
        except ValueError:
            return super()._get_weights_iterator(source)
        else:
            for name, tensor in streamer.get_tensors():
                yield source.prefix + name, tensor

    def download_model(self, model_config: ModelConfig) -> None:
        # Stream only
        pass

    def _load_sharded_model(self, *, model_config: ModelConfig, device_config: DeviceConfig) -> torch.nn.Module:
        # Forked from: https://github.com/sgl-project/sglang/blob/1c4e0d2445311f2e635e9dab5a660d982731ad20/python/sglang/srt/model_loader/loader.py#L564
        from sglang.srt.distributed import (
            get_tensor_model_parallel_rank,
            get_tensor_model_parallel_world_size,
            model_parallel_is_initialized,
        )
        from sglang.srt.model_loader.loader import (
            ShardedStateLoader,
            _initialize_model,
        )
        from sglang.srt.model_loader.utils import set_default_torch_dtype

        # Sanity checks
        if model_parallel_is_initialized():
            tensor_parallel_size = get_tensor_model_parallel_world_size()
            rank = get_tensor_model_parallel_rank()
        else:
            tensor_parallel_size = 1
            rank = 0
        if rank >= tensor_parallel_size:
            raise ValueError(f"Invalid rank {rank} for tensor parallel size {tensor_parallel_size}")
        with set_default_torch_dtype(model_config.dtype):
            with torch.device(device_config.device):
                model = _initialize_model(model_config, self.load_config)
                for _, module in model.named_modules():
                    quant_method = getattr(module, "quant_method", None)
                    if quant_method is not None:
                        quant_method.process_weights_after_loading(module)
            state_dict = ShardedStateLoader._filter_subtensors(model.state_dict())
            streamer = SafeTensorsStreamer(
                REMOTE_MODEL_PATH,
                LOCAL_MODEL_PATH,
                rank=rank,
                tensor_parallel_size=tensor_parallel_size,
            )
            for name, tensor in streamer.get_tensors():
                # If loading with LoRA enabled, additional padding may
                # be added to certain parameters. We only load into a
                # narrowed view of the parameter data.
                param_data = state_dict[name].data
                param_shape = state_dict[name].shape
                for dim, size in enumerate(tensor.shape):
                    if size < param_shape[dim]:
                        param_data = param_data.narrow(dim, 0, size)
                if tensor.shape != param_shape:
                    logger.warning(
                        "loading tensor of shape %s into parameter '%s' of shape %s",
                        tensor.shape,
                        name,
                        param_shape,
                    )
                param_data.copy_(tensor)
                state_dict.pop(name)
            if state_dict:
                raise ValueError(f"Missing keys {tuple(state_dict)} in loaded state!")
        return model.eval()

    def load_model(self, *, model_config: ModelConfig, device_config: DeviceConfig) -> torch.nn.Module:
        from sglang.srt.distributed import get_tensor_model_parallel_world_size

        if get_tensor_model_parallel_world_size() > 1:
            return self._load_sharded_model(model_config=model_config, device_config=device_config)
        else:
            return super().load_model(model_config=model_config, device_config=device_config)


# Monkeypatch the default model loader
if REMOTE_MODEL_PATH and STREAM_SAFETENSORS:
    sglang.srt.model_loader.loader.DefaultModelLoader = UnionModelLoader


def main():
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    # Prefetch the model
    if REMOTE_MODEL_PATH:
        if not prefix_exists(REMOTE_MODEL_PATH):
            raise FileNotFoundError(f"Model path not found: {REMOTE_MODEL_PATH}")

        prefetch(
            REMOTE_MODEL_PATH,
            LOCAL_MODEL_PATH,
            exclude_safetensors=STREAM_SAFETENSORS,
        )

    server_args = prepare_server_args(sys.argv[1:])
    try:
        launch_server(server_args)
    finally:
        kill_process_tree(os.getpid(), include_parent=False)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common/runtime_version.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1c\x63ommon/runtime_version.proto\x12\x0f\x63loudidl.common\"\xbd\x01\n\x0fRuntimeMetadata\x12@\n\x04type\x18\x01 \x01(\x0e\x32,.cloudidl.common.RuntimeMetadata.RuntimeTypeR\x04type\x12\x18\n\x07version\x18\x02 \x01(\tR\x07version\x12\x16\n\x06\x66lavor\x18\x03 \x01(\tR\x06\x66lavor\"6\n\x0bRuntimeType\x12\t\n\x05OTHER\x10\x00\x12\r\n\tFLYTE_SDK\x10\x01\x12\r\n\tUNION_SDK\x10\x02\x42\xb4\x01\n\x13\x63om.cloudidl.commonB\x13RuntimeVersionProtoH\x02P\x01Z)github.com/unionai/cloud/gen/pb-go/common\xa2\x02\x03\x43\x43X\xaa\x02\x0f\x43loudidl.Common\xca\x02\x0f\x43loudidl\\Common\xe2\x02\x1b\x43loudidl\\Common\\GPBMetadata\xea\x02\x10\x43loudidl::Commonb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common.runtime_version_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\023com.cloudidl.commonB\023RuntimeVersionProtoH\002P\001Z)github.com/unionai/cloud/gen/pb-go/common\242\002\003CCX\252\002\017Cloudidl.Common\312\002\017Cloudidl\\Common\342\002\033Cloudidl\\Common\\GPBMetadata\352\002\020Cloudidl::Common'
  _globals['_RUNTIMEMETADATA']._serialized_start=50
  _globals['_RUNTIMEMETADATA']._serialized_end=239
  _globals['_RUNTIMEMETADATA_RUNTIMETYPE']._serialized_start=185
  _globals['_RUNTIMEMETADATA_RUNTIMETYPE']._serialized_end=239
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hooks/hooks.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.hooks import payload_pb2 as hooks_dot_payload__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11hooks/hooks.proto\x12\x0e\x63loudidl.hooks\x1a\x13hooks/payload.proto2\xd6\x01\n\x0cHooksService\x12]\n\x0cStreamEvents\x12#.cloudidl.hooks.StreamEventsRequest\x1a$.cloudidl.hooks.StreamEventsResponse\"\x00\x30\x01\x12g\n\x10\x41\x63knowledgeEvent\x12\'.cloudidl.hooks.AcknowledgeEventRequest\x1a(.cloudidl.hooks.AcknowledgeEventResponse\"\x00\x42\xa5\x01\n\x12\x63om.cloudidl.hooksB\nHooksProtoH\x02P\x01Z(github.com/unionai/cloud/gen/pb-go/hooks\xa2\x02\x03\x43HX\xaa\x02\x0e\x43loudidl.Hooks\xca\x02\x0e\x43loudidl\\Hooks\xe2\x02\x1a\x43loudidl\\Hooks\\GPBMetadata\xea\x02\x0f\x43loudidl::Hooksb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'hooks.hooks_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\022com.cloudidl.hooksB\nHooksProtoH\002P\001Z(github.com/unionai/cloud/gen/pb-go/hooks\242\002\003CHX\252\002\016Cloudidl.Hooks\312\002\016Cloudidl\\Hooks\342\002\032Cloudidl\\Hooks\\GPBMetadata\352\002\017Cloudidl::Hooks'
  _globals['_HOOKSSERVICE']._serialized_start=59
  _globals['_HOOKSSERVICE']._serialized_end=273
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import typing
from dataclasses import dataclass, field

import rich.repr
from flytekit.core.artifact import Partition
from flytekit.extras import accelerators as acc

from union import Artifact, Resources
from union.app import App, Input

ARCHITECTURE_KEY = "architecture"
TASK_KEY = "task"
FORMAT_KEY = "format"
MODALITY_KEY = "modality"
ARTIFACT_TYPE_KEY = "_u_type"
MODEL_TYPE_KEY = "model_type"
HUGGINGFACE_SOURCE_KEY = "huggingface-source"
SHARD_ENGINE_KEY = "shard_engine"
SHARD_PARALLELISM_KEY = "shard_parallelism"
COMMIT_KEY = "commit"

_FASTAPI_ARCHITECTURES = ["xgboost".casefold(), "custom".casefold(), "joblib".casefold(), "sklearn".casefold()]

VLLM_IMAGE = "ghcr.io/unionai-oss/serving-vllm:0.1.16"


@dataclass
@rich.repr.auto()
class AppTemplate:
    app: App
    engine: str
    preferred: bool = False


def _sanitize_model_name(name: str) -> str:
    return name.lower().replace("/", "-").replace(":", "-").replace(".", "-").replace("_", "-")[:25]


@dataclass
class AppTemplateParams:
    limits: Resources
    accelerator: acc.GPUAccelerator
    prefrerred: bool = False


_TEMPLATES: typing.Dict[str, AppTemplateParams] = {
    "qwen2": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "llama3": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "llama2": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "llama": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "mistral": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "gemma": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "mixtral": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "falcon": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "qwen": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
    "stablelm": AppTemplateParams(
        limits=Resources(cpu="7", mem="27Gi", gpu="1", ephemeral_storage="10Gi"),
        accelerator=acc.L4,
        prefrerred=True,
    ),
}


class VLLMApp(App):
    def __init__(
        self,
        name: str,
        model_input: typing.Union[Artifact, str],
        accelerator: acc.GPUAccelerator,
        limits: Resources,
        requires_auth: bool = False,
        **kwargs,
    ):
        if isinstance(model_input, Artifact):
            raise NotImplementedError("Artifact input not supported yet")
        super().__init__(
            name=name,
            container_image=VLLM_IMAGE,
            args=f"union-model-loader-vllm serve /home/union --served-model-name {name}",
            port=8000,
            limits=limits,
            accelerator=accelerator,
            env={
                "UNION_MODEL_LOADER_STREAM_SAFETENSORS": "true",
                "UNION_MODEL_LOADER_LOCAL_MODEL_PATH": "/home/union",
                "UNION_MODEL_LOADER_REMOTE_MODEL_PATH": model_input,
            },
            requires_auth=requires_auth,
        )


def get_vllm_app_template(model: Artifact) -> AppTemplate:
    if model.literal is None or model.literal.scalar is None or model.literal.scalar.blob is None:
        raise ValueError(f"Model artifact must have a Directory or a File location, found {model.literal}")
    model_uri = model.literal.scalar.blob.uri
    model_type = find_model_type_for_model(model)
    if model_type is None:
        raise ValueError("Model type not found")
    v = model_type.value.static_value
    if v not in _TEMPLATES:
        raise ValueError(f"Model type {model_type.value} not supported")
    template = _TEMPLATES[v]
    return AppTemplate(
        app=VLLMApp(
            name=f"{_sanitize_model_name(model.name)}-vllm",
            model_input=model_uri,
            accelerator=template.accelerator,
            limits=template.limits,
        ),
        engine="vllm",
        preferred=template.prefrerred,
    )


def _get_fastapi_app_template(model: Artifact) -> AppTemplate:
    return AppTemplate(
        app=App(
            name=f"{_sanitize_model_name(model.name)}-fapi",
            container_image="",
            inputs=[
                Input(
                    name="model",
                    value=model.query(),
                    env_var="MODEL_FILE_PATH",
                ),
            ],
            limits=Resources(cpu="2", mem="4Gi"),
            port=8080,
            include=["./app.py"],
            command=["uvicorn", "--port", "8080", "app:app"],
        ),
        engine="fastapi",
        preferred=True,
    )


def get_partition_keys_for_model(
    architecture: str,
    task: str,
    modality: typing.List[str] | None = None,
    serial_format: str = "safetensors",
    model_type: str | None = None,
) -> typing.Dict[str, str]:
    """
    Get partition keys for a model, given the architecture, task, modality, serial format and model type.
    :param architecture: The architecture of the model
    :param task: The task of the model
    :param modality: The modality of the model
    :param serial_format: The serial format of the model
    :param model_type: The model type of the model
    :return: The partition keys for the model
    """
    if not modality:
        modality = ["text"]

    if architecture is None:
        architecture = "custom"

    if task is None:
        task = "auto"

    if model_type is None:
        model_type = "custom"

    return {
        ARCHITECTURE_KEY: architecture,
        TASK_KEY: task,
        FORMAT_KEY: serial_format,
        MODALITY_KEY: ",".join(modality),
        ARTIFACT_TYPE_KEY: "model",
        MODEL_TYPE_KEY: model_type,
    }


def find_architecture_for_model(model: Artifact) -> Partition | None:
    if model.partitions and model.partitions.partitions:
        return model.partitions.partitions.get(ARCHITECTURE_KEY, None)
    return None


def find_model_type_for_model(model: Artifact) -> Partition | None:
    if model.partitions and model.partitions.partitions:
        return model.partitions.partitions.get(MODEL_TYPE_KEY, None)
    return None


def get_app_templates_for_model(model: Artifact) -> typing.List[AppTemplate]:
    """Get app templates for a model."""
    mt = find_model_type_for_model(model)
    if mt and mt.value:
        # TODO: Update when xgboost is actually supported
        if mt.value.static_value.casefold() == "xgboost":
            return [_get_fastapi_app_template(model)]
    return [get_vllm_app_template(model)]


@dataclass
@rich.repr.auto()
class ShardConfig:
    engine: str
    args: VLLMShardArgs

    def __post_init__(self):
        if self.engine != "vllm":
            raise ValueError(f"Unsupported engine: {self.engine}")


@dataclass
@rich.repr.auto()
class VLLMShardArgs:
    model: str
    tensor_parallel_size: int
    trust_remote_code: bool = False
    revision: str | None = None
    file_pattern: str | None = None  # string pattern of saved filenames
    max_file_size: int | None = 5 * 1024**3  # max size (in bytes) of each safetensors file
    gpu_memory_utilization: float = 0.9
    # extra arguments to pass to the vllm.EngineArgs
    # https://docs.vllm.ai/en/stable/api/vllm/engine/arg_utils.html#vllm.engine.arg_utils.EngineArgs
    extra_args: dict[str, typing.Any] = field(default_factory=dict)

    def get_vllm_args(self, model_path: str):
        return {
            "model": model_path,
            "tensor_parallel_size": self.tensor_parallel_size,
            "trust_remote_code": self.trust_remote_code,
            "revision": self.revision,
            "gpu_memory_utilization": self.gpu_memory_utilization,
            **self.extra_args,
        }


@dataclass
@rich.repr.auto()
class HuggingFaceModelInfo:
    """
    Captures information about a Hugging Face model. Only repo is required, all other fields are optional, and are
    automatically determined from the model's config.json file. If not found, the fields are initialized to defaults.

    :param repo: The model repo name in huggingface.
    :param artifact_name: The name of the Union artifact to use for the cached model.
    :param model_type: The model type.
    :param architecture: The model architecture.
    :param task: The model task.
    :param modality: The model modality.
    :param serial_format: The model serialization format.
    :param short_description: A short description of the model.
    :param shard_config: Configuration to shard the model with.
    """

    repo: str
    artifact_name: str | None = None
    model_type: str | None = None
    architecture: str | None = None
    task: str = "auto"
    modality: typing.List[str] | None = None
    serial_format: str = "safetensors"
    short_description: str | None = None
    shard_config: ShardConfig | None = None

    def __post_init__(self):
        if self.modality is None:
            self.modality = ["text"]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from union.internal.workspace import workspace_definition_payload_pb2 as workspace_dot_workspace__definition__payload__pb2


class WorkspaceRegistryServiceStub(object):
    """WorkspaceRegistryService defines the inventory service for creating and fetching workspace definitions.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateWorkspaceDefinition = channel.unary_unary(
                '/cloudidl.workspace.WorkspaceRegistryService/CreateWorkspaceDefinition',
                request_serializer=workspace_dot_workspace__definition__payload__pb2.CreateWorkspaceDefinitionRequest.SerializeToString,
                response_deserializer=workspace_dot_workspace__definition__payload__pb2.CreateWorkspaceDefinitionResponse.FromString,
                )
        self.GetWorkspaceDefinition = channel.unary_unary(
                '/cloudidl.workspace.WorkspaceRegistryService/GetWorkspaceDefinition',
                request_serializer=workspace_dot_workspace__definition__payload__pb2.GetWorkspaceDefinitionRequest.SerializeToString,
                response_deserializer=workspace_dot_workspace__definition__payload__pb2.GetWorkspaceDefinitionResponse.FromString,
                )
        self.UpdateWorkspaceDefinitionEntity = channel.unary_unary(
                '/cloudidl.workspace.WorkspaceRegistryService/UpdateWorkspaceDefinitionEntity',
                request_serializer=workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionEntityRequest.SerializeToString,
                response_deserializer=workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionEntityResponse.FromString,
                )
        self.UpdateWorkspaceDefinition = channel.unary_unary(
                '/cloudidl.workspace.WorkspaceRegistryService/UpdateWorkspaceDefinition',
                request_serializer=workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionRequest.SerializeToString,
                response_deserializer=workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionResponse.FromString,
                )
        self.ListWorkspaceDefinitionEntities = channel.unary_unary(
                '/cloudidl.workspace.WorkspaceRegistryService/ListWorkspaceDefinitionEntities',
                request_serializer=workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionEntitiesRequest.SerializeToString,
                response_deserializer=workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionEntitiesResponse.FromString,
                )
        self.ListWorkspaceDefinitions = channel.unary_unary(
                '/cloudidl.workspace.WorkspaceRegistryService/ListWorkspaceDefinitions',
                request_serializer=workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionsRequest.SerializeToString,
                response_deserializer=workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionsResponse.FromString,
                )


class WorkspaceRegistryServiceServicer(object):
    """WorkspaceRegistryService defines the inventory service for creating and fetching workspace definitions.
    """

    def CreateWorkspaceDefinition(self, request, context):
        """CreateWorkspaceDefinition creates a new workspace definition.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetWorkspaceDefinition(self, request, context):
        """GetWorkspaceDefinition fetches a workspace definition by its ID.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateWorkspaceDefinitionEntity(self, request, context):
        """UpdateWorkspaceDefinitionEntity updates an existing workspace definition entity.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateWorkspaceDefinition(self, request, context):
        """UpdateWorkspaceDefinition updates an existing workspace definition.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListWorkspaceDefinitionEntities(self, request, context):
        """ListWorkspaceDefinitionEntities fetches all workspace definition entities for a given organization and project.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListWorkspaceDefinitions(self, request, context):
        """ListWorkspaceDefinitions fetches all workspace definitions for a given organization and project.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WorkspaceRegistryServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateWorkspaceDefinition': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateWorkspaceDefinition,
                    request_deserializer=workspace_dot_workspace__definition__payload__pb2.CreateWorkspaceDefinitionRequest.FromString,
                    response_serializer=workspace_dot_workspace__definition__payload__pb2.CreateWorkspaceDefinitionResponse.SerializeToString,
            ),
            'GetWorkspaceDefinition': grpc.unary_unary_rpc_method_handler(
                    servicer.GetWorkspaceDefinition,
                    request_deserializer=workspace_dot_workspace__definition__payload__pb2.GetWorkspaceDefinitionRequest.FromString,
                    response_serializer=workspace_dot_workspace__definition__payload__pb2.GetWorkspaceDefinitionResponse.SerializeToString,
            ),
            'UpdateWorkspaceDefinitionEntity': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateWorkspaceDefinitionEntity,
                    request_deserializer=workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionEntityRequest.FromString,
                    response_serializer=workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionEntityResponse.SerializeToString,
            ),
            'UpdateWorkspaceDefinition': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateWorkspaceDefinition,
                    request_deserializer=workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionRequest.FromString,
                    response_serializer=workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionResponse.SerializeToString,
            ),
            'ListWorkspaceDefinitionEntities': grpc.unary_unary_rpc_method_handler(
                    servicer.ListWorkspaceDefinitionEntities,
                    request_deserializer=workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionEntitiesRequest.FromString,
                    response_serializer=workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionEntitiesResponse.SerializeToString,
            ),
            'ListWorkspaceDefinitions': grpc.unary_unary_rpc_method_handler(
                    servicer.ListWorkspaceDefinitions,
                    request_deserializer=workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionsRequest.FromString,
                    response_serializer=workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cloudidl.workspace.WorkspaceRegistryService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class WorkspaceRegistryService(object):
    """WorkspaceRegistryService defines the inventory service for creating and fetching workspace definitions.
    """

    @staticmethod
    def CreateWorkspaceDefinition(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.workspace.WorkspaceRegistryService/CreateWorkspaceDefinition',
            workspace_dot_workspace__definition__payload__pb2.CreateWorkspaceDefinitionRequest.SerializeToString,
            workspace_dot_workspace__definition__payload__pb2.CreateWorkspaceDefinitionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetWorkspaceDefinition(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.workspace.WorkspaceRegistryService/GetWorkspaceDefinition',
            workspace_dot_workspace__definition__payload__pb2.GetWorkspaceDefinitionRequest.SerializeToString,
            workspace_dot_workspace__definition__payload__pb2.GetWorkspaceDefinitionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateWorkspaceDefinitionEntity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.workspace.WorkspaceRegistryService/UpdateWorkspaceDefinitionEntity',
            workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionEntityRequest.SerializeToString,
            workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionEntityResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateWorkspaceDefinition(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.workspace.WorkspaceRegistryService/UpdateWorkspaceDefinition',
            workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionRequest.SerializeToString,
            workspace_dot_workspace__definition__payload__pb2.UpdateWorkspaceDefinitionResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListWorkspaceDefinitionEntities(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.workspace.WorkspaceRegistryService/ListWorkspaceDefinitionEntities',
            workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionEntitiesRequest.SerializeToString,
            workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionEntitiesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListWorkspaceDefinitions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.workspace.WorkspaceRegistryService/ListWorkspaceDefinitions',
            workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionsRequest.SerializeToString,
            workspace_dot_workspace__definition__payload__pb2.ListWorkspaceDefinitionsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from union.internal.objectstore import payload_pb2 as objectstore_dot_payload__pb2


class ObjectStoreServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Metadata = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/Metadata',
                request_serializer=objectstore_dot_payload__pb2.MetadataRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.MetadataResponse.FromString,
                )
        self.Head = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/Head',
                request_serializer=objectstore_dot_payload__pb2.HeadRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.HeadResponse.FromString,
                )
        self.Put = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/Put',
                request_serializer=objectstore_dot_payload__pb2.PutRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.PutResponse.FromString,
                )
        self.Get = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/Get',
                request_serializer=objectstore_dot_payload__pb2.GetRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.GetResponse.FromString,
                )
        self.Delete = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/Delete',
                request_serializer=objectstore_dot_payload__pb2.DeleteRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.DeleteResponse.FromString,
                )
        self.List = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/List',
                request_serializer=objectstore_dot_payload__pb2.ListRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.ListResponse.FromString,
                )
        self.Copy = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/Copy',
                request_serializer=objectstore_dot_payload__pb2.CopyRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.CopyResponse.FromString,
                )
        self.StartMultipartUpload = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/StartMultipartUpload',
                request_serializer=objectstore_dot_payload__pb2.StartMultipartUploadRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.StartMultipartUploadResponse.FromString,
                )
        self.UploadPart = channel.stream_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/UploadPart',
                request_serializer=objectstore_dot_payload__pb2.UploadPartRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.UploadPartResponse.FromString,
                )
        self.ListInProgressMultipartUploads = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/ListInProgressMultipartUploads',
                request_serializer=objectstore_dot_payload__pb2.ListInProgressMultipartUploadsRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.ListInProgressMultipartUploadsResponse.FromString,
                )
        self.TerminateMultipartUpload = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/TerminateMultipartUpload',
                request_serializer=objectstore_dot_payload__pb2.TerminateMultipartUploadRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.TerminateMultipartUploadResponse.FromString,
                )
        self.DownloadPart = channel.unary_stream(
                '/cloudidl.objectstore.v1.ObjectStoreService/DownloadPart',
                request_serializer=objectstore_dot_payload__pb2.DownloadPartRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.DownloadPartResponse.FromString,
                )
        self.Presign = channel.unary_unary(
                '/cloudidl.objectstore.v1.ObjectStoreService/Presign',
                request_serializer=objectstore_dot_payload__pb2.PresignRequest.SerializeToString,
                response_deserializer=objectstore_dot_payload__pb2.PresignResponse.FromString,
                )


class ObjectStoreServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Metadata(self, request, context):
        """Metadata returns the metadata for the object store
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Head(self, request, context):
        """Head returns the metadata for an object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Put(self, request, context):
        """Put creates or updates an object. The maximum size of an object is 1GB.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Get(self, request, context):
        """Get returns the contents of an object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Delete(self, request, context):
        """Delete deletes an object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def List(self, request, context):
        """List lists the objects
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Copy(self, request, context):
        """Copy copies an object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StartMultipartUpload(self, request, context):
        """StartMultipartUpload starts a large object upload. A large object is any object whose size is greater than
        metadata.max_single_part_object_size_bytes.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UploadPart(self, request_iterator, context):
        """UploadPart uploads a part of a large object
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListInProgressMultipartUploads(self, request, context):
        """ListInProgressMultipartUploads completes a large object upload
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TerminateMultipartUpload(self, request, context):
        """TerminateMultipartUpload terminates a large object upload
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DownloadPart(self, request, context):
        """DownloadPart downloads a part of a large object.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Presign(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ObjectStoreServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Metadata': grpc.unary_unary_rpc_method_handler(
                    servicer.Metadata,
                    request_deserializer=objectstore_dot_payload__pb2.MetadataRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.MetadataResponse.SerializeToString,
            ),
            'Head': grpc.unary_unary_rpc_method_handler(
                    servicer.Head,
                    request_deserializer=objectstore_dot_payload__pb2.HeadRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.HeadResponse.SerializeToString,
            ),
            'Put': grpc.unary_unary_rpc_method_handler(
                    servicer.Put,
                    request_deserializer=objectstore_dot_payload__pb2.PutRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.PutResponse.SerializeToString,
            ),
            'Get': grpc.unary_unary_rpc_method_handler(
                    servicer.Get,
                    request_deserializer=objectstore_dot_payload__pb2.GetRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.GetResponse.SerializeToString,
            ),
            'Delete': grpc.unary_unary_rpc_method_handler(
                    servicer.Delete,
                    request_deserializer=objectstore_dot_payload__pb2.DeleteRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.DeleteResponse.SerializeToString,
            ),
            'List': grpc.unary_unary_rpc_method_handler(
                    servicer.List,
                    request_deserializer=objectstore_dot_payload__pb2.ListRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.ListResponse.SerializeToString,
            ),
            'Copy': grpc.unary_unary_rpc_method_handler(
                    servicer.Copy,
                    request_deserializer=objectstore_dot_payload__pb2.CopyRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.CopyResponse.SerializeToString,
            ),
            'StartMultipartUpload': grpc.unary_unary_rpc_method_handler(
                    servicer.StartMultipartUpload,
                    request_deserializer=objectstore_dot_payload__pb2.StartMultipartUploadRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.StartMultipartUploadResponse.SerializeToString,
            ),
            'UploadPart': grpc.stream_unary_rpc_method_handler(
                    servicer.UploadPart,
                    request_deserializer=objectstore_dot_payload__pb2.UploadPartRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.UploadPartResponse.SerializeToString,
            ),
            'ListInProgressMultipartUploads': grpc.unary_unary_rpc_method_handler(
                    servicer.ListInProgressMultipartUploads,
                    request_deserializer=objectstore_dot_payload__pb2.ListInProgressMultipartUploadsRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.ListInProgressMultipartUploadsResponse.SerializeToString,
            ),
            'TerminateMultipartUpload': grpc.unary_unary_rpc_method_handler(
                    servicer.TerminateMultipartUpload,
                    request_deserializer=objectstore_dot_payload__pb2.TerminateMultipartUploadRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.TerminateMultipartUploadResponse.SerializeToString,
            ),
            'DownloadPart': grpc.unary_stream_rpc_method_handler(
                    servicer.DownloadPart,
                    request_deserializer=objectstore_dot_payload__pb2.DownloadPartRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.DownloadPartResponse.SerializeToString,
            ),
            'Presign': grpc.unary_unary_rpc_method_handler(
                    servicer.Presign,
                    request_deserializer=objectstore_dot_payload__pb2.PresignRequest.FromString,
                    response_serializer=objectstore_dot_payload__pb2.PresignResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cloudidl.objectstore.v1.ObjectStoreService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ObjectStoreService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Metadata(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/Metadata',
            objectstore_dot_payload__pb2.MetadataRequest.SerializeToString,
            objectstore_dot_payload__pb2.MetadataResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Head(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/Head',
            objectstore_dot_payload__pb2.HeadRequest.SerializeToString,
            objectstore_dot_payload__pb2.HeadResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Put(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/Put',
            objectstore_dot_payload__pb2.PutRequest.SerializeToString,
            objectstore_dot_payload__pb2.PutResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Get(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/Get',
            objectstore_dot_payload__pb2.GetRequest.SerializeToString,
            objectstore_dot_payload__pb2.GetResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Delete(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/Delete',
            objectstore_dot_payload__pb2.DeleteRequest.SerializeToString,
            objectstore_dot_payload__pb2.DeleteResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def List(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/List',
            objectstore_dot_payload__pb2.ListRequest.SerializeToString,
            objectstore_dot_payload__pb2.ListResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Copy(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/Copy',
            objectstore_dot_payload__pb2.CopyRequest.SerializeToString,
            objectstore_dot_payload__pb2.CopyResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StartMultipartUpload(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/StartMultipartUpload',
            objectstore_dot_payload__pb2.StartMultipartUploadRequest.SerializeToString,
            objectstore_dot_payload__pb2.StartMultipartUploadResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UploadPart(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/cloudidl.objectstore.v1.ObjectStoreService/UploadPart',
            objectstore_dot_payload__pb2.UploadPartRequest.SerializeToString,
            objectstore_dot_payload__pb2.UploadPartResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListInProgressMultipartUploads(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/ListInProgressMultipartUploads',
            objectstore_dot_payload__pb2.ListInProgressMultipartUploadsRequest.SerializeToString,
            objectstore_dot_payload__pb2.ListInProgressMultipartUploadsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TerminateMultipartUpload(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/TerminateMultipartUpload',
            objectstore_dot_payload__pb2.TerminateMultipartUploadRequest.SerializeToString,
            objectstore_dot_payload__pb2.TerminateMultipartUploadResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DownloadPart(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/DownloadPart',
            objectstore_dot_payload__pb2.DownloadPartRequest.SerializeToString,
            objectstore_dot_payload__pb2.DownloadPartResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Presign(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cloudidl.objectstore.v1.ObjectStoreService/Presign',
            objectstore_dot_payload__pb2.PresignRequest.SerializeToString,
            objectstore_dot_payload__pb2.PresignResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from dataclasses import dataclass
from typing import TYPE_CHECKING

from ._models import AppConfigProtocol, AppSerializationSettings, Link

if TYPE_CHECKING:
    from ._models import App


@dataclass
class WeaveConfig(AppConfigProtocol):
    project: str
    entity: str
    api_host: str = "https://api.wandb.ai"
    host: str = "https://wandb.ai"

    def before_to_union_idl(self, app: "App", settings: AppSerializationSettings):
        """Modify app in place at the beginning of `App._to_union_idl`."""
        app.env["WANDB_BASE_URL"] = self.api_host
        app.env["WANDB_PROJECT"] = self.project

        weave_url = f"{self.host.rstrip('/')}/{self.entity}/{self.project}/weave/traces"

        link = Link(path=weave_url, title="Weave Traces")
        app.links.append(link)



from union.internal.common import authorization_pb2 as _authorization_pb2
from union.internal.common import identifier_pb2 as _identifier_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RoleType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    ROLE_TYPE_NONE: _ClassVar[RoleType]
    ROLE_TYPE_ADMIN: _ClassVar[RoleType]
    ROLE_TYPE_CONTRIBUTOR: _ClassVar[RoleType]
    ROLE_TYPE_VIEWER: _ClassVar[RoleType]
    ROLE_TYPE_CUSTOM: _ClassVar[RoleType]
    ROLE_TYPE_CLUSTER_MANAGER: _ClassVar[RoleType]
    ROLE_TYPE_FLYTE_PROJECT_ADMIN: _ClassVar[RoleType]
    ROLE_TYPE_SERVERLESS_VIEWER: _ClassVar[RoleType]
    ROLE_TYPE_SERVERLESS_CONTRIBUTOR: _ClassVar[RoleType]
ROLE_TYPE_NONE: RoleType
ROLE_TYPE_ADMIN: RoleType
ROLE_TYPE_CONTRIBUTOR: RoleType
ROLE_TYPE_VIEWER: RoleType
ROLE_TYPE_CUSTOM: RoleType
ROLE_TYPE_CLUSTER_MANAGER: RoleType
ROLE_TYPE_FLYTE_PROJECT_ADMIN: RoleType
ROLE_TYPE_SERVERLESS_VIEWER: RoleType
ROLE_TYPE_SERVERLESS_CONTRIBUTOR: RoleType

class Role(_message.Message):
    __slots__ = ["id", "permissions", "role_spec", "role_type", "actions"]
    ID_FIELD_NUMBER: _ClassVar[int]
    PERMISSIONS_FIELD_NUMBER: _ClassVar[int]
    ROLE_SPEC_FIELD_NUMBER: _ClassVar[int]
    ROLE_TYPE_FIELD_NUMBER: _ClassVar[int]
    ACTIONS_FIELD_NUMBER: _ClassVar[int]
    id: _identifier_pb2.RoleIdentifier
    permissions: _containers.RepeatedCompositeFieldContainer[_authorization_pb2.Permission]
    role_spec: RoleSpec
    role_type: RoleType
    actions: _containers.RepeatedScalarFieldContainer[_authorization_pb2.Action]
    def __init__(self, id: _Optional[_Union[_identifier_pb2.RoleIdentifier, _Mapping]] = ..., permissions: _Optional[_Iterable[_Union[_authorization_pb2.Permission, _Mapping]]] = ..., role_spec: _Optional[_Union[RoleSpec, _Mapping]] = ..., role_type: _Optional[_Union[RoleType, str]] = ..., actions: _Optional[_Iterable[_Union[_authorization_pb2.Action, str]]] = ...) -> None: ...

class RoleSpec(_message.Message):
    __slots__ = ["description"]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    description: str
    def __init__(self, description: _Optional[str] = ...) -> None: ...



"""
Vendored from https://github.com/gweis/isodate

Copyright (c) 2021, Hugo van Kemenade and contributors
Copyright (c) 2009-2018, Gerhard Weis and contributors
Copyright (c) 2009, Gerhard Weis
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""

import re
from datetime import timedelta

STRF_D_MAP = {
    "%d": lambda tdt, yds: "%02d" % tdt.days,
    "%f": lambda tdt, yds: "%06d" % tdt.microseconds,
    "%H": lambda tdt, yds: "%02d" % (tdt.seconds / 60 / 60),
    "%m": lambda tdt, yds: "%02d" % tdt.months,
    "%M": lambda tdt, yds: "%02d" % ((tdt.seconds / 60) % 60),
    "%S": lambda tdt, yds: "%02d" % (tdt.seconds % 60),
    "%W": lambda tdt, yds: "%02d" % (abs(tdt.days / 7)),
    "%Y": lambda tdt, yds: (((yds != 4) and "+") or "") + (("%%0%dd" % yds) % tdt.years),
    "%C": lambda tdt, yds: (((yds != 4) and "+") or "") + (("%%0%dd" % (yds - 2)) % (tdt.years / 100)),
    "%%": lambda tdt, yds: "%",
}


def _strfduration(tdt, format, yeardigits=4):
    """
    this is the work method for timedelta and Duration instances.

    see strftime for more details.
    """

    def repl(match):
        """
        lookup format command and return corresponding replacement.
        """
        if match.group(0) in STRF_D_MAP:
            return STRF_D_MAP[match.group(0)](tdt, yeardigits)
        elif match.group(0) == "%P":
            ret = []
            usecs = abs((tdt.days * 24 * 60 * 60 + tdt.seconds) * 1000000 + tdt.microseconds)
            seconds, usecs = divmod(usecs, 1000000)
            minutes, seconds = divmod(seconds, 60)
            hours, minutes = divmod(minutes, 60)
            days, hours = divmod(hours, 24)
            if days:
                ret.append("%sD" % days)
            if hours or minutes or seconds or usecs:
                ret.append("T")
                if hours:
                    ret.append("%sH" % hours)
                if minutes:
                    ret.append("%sM" % minutes)
                if seconds or usecs:
                    if usecs:
                        ret.append(("%d.%06d" % (seconds, usecs)).rstrip("0"))
                    else:
                        ret.append("%d" % seconds)
                    ret.append("S")
            # at least one component has to be there.
            return (ret and "".join(ret)) or "0D"
        elif match.group(0) == "%p":
            return str(abs(tdt.days // 7)) + "W"
        return match.group(0)

    return re.sub("%d|%f|%H|%m|%M|%S|%W|%Y|%C|%%|%P|%p", repl, format)


D_DEFAULT = "P%P"


def duration_isoformat(tduration, format=D_DEFAULT):
    """
    Format duration strings.

    This method is just a wrapper around isodate.isostrf.strftime and uses
    P%P (D_DEFAULT) as default format.
    """
    # TODO: implement better decision for negative Durations.
    #       should be done in Duration class in consistent way with timedelta.
    if isinstance(tduration, timedelta) and (tduration < timedelta(0)):
        ret = "-"
    else:
        ret = ""
    ret += _strfduration(tduration, format)
    return ret
